!+
! Module: esrcselect_mod
!
! Authors:  Richard West
!          
! Description: Select EPIC sources for source product generation
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module esrcselect_mod

  use dal
  use param  
  use errorHandling
  use sort_mod
  
  implicit none 
  
  private 

  public esrcselect

contains 
  
  !---------------------------------------------------------------------
  subroutine esrcselect
  !---------------------------------------------------------------------
    character (len=1024) :: obsList, expList, setName, outList
    character (len=10) :: opmode, instrument, attName,&
         & string
    type (DataSetT) :: obsSet, expSet, outSet
    type (TableT) :: obsTable, expTable, outTable
    type (ColumnT) :: obsRateCol, obsIDCol, srcRACol, srcDecCol,&
         & srcCtCol, nnCol, listXIDCol, bandCol, instCol, expIDCol,&
         & outIDCol, outRACol, outDecCol, outRateCol, outCountsCol,&
         & outDistCol, flagCol
    real :: minRate, minCounts, minSeperation, maxRate
    integer :: maxSources, nObsSrc, nExpSrc, nOutSrc, instID, i, j,&
         & setNum
    integer, pointer, dimension(:) :: expID, outID, srcExpIndex,&
         & expInst, expBand, obsID
    integer, allocatable, dimension(:) :: tempID, sortIndex
    integer, pointer, dimension(:,:) :: listXID
    real (kind=double), pointer, dimension(:) :: srcRA, srcDec, outRA,&
         & outDec
    real (kind=double), allocatable, dimension(:) :: tempRA, tempDec
    real, allocatable, dimension(:) :: tempRate, tempCounts, tempDist
    real, pointer, dimension(:) :: countRate, srcCounts, nnDist, outRate,&
         & outCounts, outDist
    logical :: more, ok
    logical (kind=boolean), pointer, dimension(:) :: flag

    ! Temp
    integer, allocatable, dimension(:) :: tmpIdx
    real, allocatable, dimension(:) :: tmpTmp

    character (len=2) :: instCode(3)=(/'PN','M1','M2'/)

    ! Variables are:
    ! expID:     source number from exposure list (ML_ID_SRC)
    ! expInst:   instrument ID from exposure list (ID_INST)
    ! expBand:   band ID from exposure list (ID_BAND)
    ! listXID:   vector column, from obs. list, cross-linking observation
    !            source IDs to those in exposure lists
    ! obsID:     source number from observation list (SRC_NUM)
    ! outID, outRA, outDec : output vectors

    ! Read parameters
    obsList=stringParameter('obssourcelist')
    expList=stringParameter('expsourcelist')
    opmode=stringParameter('operationmode')
    minRate=realParameter('minsourcerate')
    maxRate=realParameter('maxsourcerate')
    minCounts=realParameter('minsourcecounts')
    minSeperation=realParameter('minsourceseparation')
    maxSources=intParameter('maxsources')
    outList=stringParameter('outsourcelist')

    ! Open observation source list
    obsSet=dataSet(obsList, Modify)
    obsTable=table(obsSet, 'SRCLIST')
    nObsSrc=numberOfRows(obsTable)

    ! Open exposure source list, read instrument keyword
    expSet=dataSet(expList, Read)
    expTable=table(expSet, 'SRCLIST')
    instrument=stringAttribute(expSet, "INSTRUME")
    nExpSrc=numberOfRows(expTable)

    ! Read observation source ID, count rate, nearest neighbour distance,
    ! and exposure <-> observation cross-ID from observation source list
    obsIDCol=column(obsTable, "SRC_NUM", Read)
    if (instrument == 'EPN') then
       obsRateCol=column(obsTable, "PN_TOT", Read)
       listXIDCol=column(obsTable, "PN_ML_ID_SRC", Read)
       instID=1
    else if(instrument == 'EMOS1') then
       obsRateCol=column(obsTable, "M1_TOT", Read)
       listXIDCol=column(obsTable, "M1_ML_ID_SRC", Read)
       instID=2
    else if(instrument == 'EMOS2') then
       obsRateCol=column(obsTable, "M2_TOT", Read)
       listXIDCol=column(obsTable, "M2_ML_ID_SRC", Read)
       instID=3
    else
       call error('badInstrument',&
            & 'Unrecognised instrument identifier '''//trim(instrument)//'''')
    end if
    
    ! Fetch data
    obsID=>int32Data(obsIDCol)
    listXID=>int32VectorData(listXIDCol)
    countRate=>real32Data(obsRateCol)

    ! Fetch flag column (creating it if it doesn't exist)
    if(opmode == 'spectrum') then
       string='SPECTRA'
    else
       string='TIMESERIES'
    end if

    if(hasColumn(obsTable, string)) then
       flagCol=column(obsTable, string, MODIFY)
    else
       flagCol=addColumn(obsTable, string, BOOLEAN)
    end if
    flag=>boolData(flagCol)
    
    ! Read RA, Dec., source counts, band, instrument from
    ! exposure source list
    expIDCol=column(expTable, "ML_ID_SRC", Read)
    srcCtCol=column(expTable, "SCTS", Read)
    srcRACol=column(expTable, "RA", Read)
    srcDecCol=column(expTable, "DEC", Read)
    instCol=column(expTable, "ID_INST", Read)
    bandCol=column(expTable, "ID_BAND", Read)
    nnCol=column(expTable, "DIST_NN", Read)

    ! Fetch data
    expID=>int32Data(expIDCol)
    srcCounts=>real32Data(srcCtCol)
    srcRA=>real64Data(srcRACol)
    srcDec=>real64Data(srcDecCol)
    expBand=>int32Data(bandCol)
    expInst=>int32Data(instCol)
    nnDist=>real32Data(nnCol)

    ! Search through the attributes of the table to work out which
    ! set number the supplied exposure list corresponds to. This must
    ! be done by comparing filenames (bleugh!)
    setNum=-1
    expList=expList(index(expList,'/',back=.true.)+1:)
    do i=0,size(listXID,1)-1
       write(string,fmt='(I3)')i+1 ! Because we "love" FORTRAN, we have to
       attName=instCode(instID)//'_SET'//adjustl(string) ! do all this

       setName=stringAttribute(obsTable, attName)
       setName=setName(index(setName,'/',back=.true.)+1:)
       if(setName == expList) setNum=i
    end do
    if(setNum==-1)call error('setNotFound',&
         & 'Cannot find reference to exposure list in observation list')

    ! For each source in the observation list, work out the index of
    ! the source in the exposure list (for the given instrument, and
    ! full energy band)
    allocate(srcExpIndex(0:nObsSrc-1))
    srcExpIndex=-1
    do i=0, nObsSrc-1
       do j=0, nExpSrc-1
          if(expID(j)==listXID(setNum, i) .and. expBand(j)==0 .and. &
               & expInst(j)==instID) srcExpIndex(i)=j
       end do
    end do

    ! Sort observation list into descending order of count rate
    allocate(tempRate(0:nObsSrc-1))
    allocate(sortIndex(0:nObsSrc-1))
    tempRate=countRate

    ! Work round bug in quick_sort (assumes arrays start at 1)
    allocate(tmpIdx(nObsSrc))
    allocate(tmpTmp(nObsSrc))
    tmpTmp=tempRate
    call quick_sort(tmpTmp, tmpIdx)
    sortIndex=tmpIdx-1
    deallocate(tmpTmp)
    deallocate(tmpIdx)

    ! Allocate space for output list
    allocate(tempID(0:nObsSrc-1))
    allocate(tempRA(0:nObsSrc-1))
    allocate(tempDec(0:nObsSrc-1))
    allocate(tempCounts(0:nObsSrc-1))
    allocate(tempDist(0:nObsSrc-1))
    
    ! Step through list, selecting sources
    nOutSrc=-1
    i=nObsSrc-1
    more=.true.
    flag=.false.
    do while (more)
       ok=.true.

       ! Index into sorted list
       j=sortIndex(i)

       ! Check source exists in exposure
       ok=(srcExpIndex(j)/=-1)

       ! Check count rate
       if(ok) ok=(countRate(j) > minRate .and. countRate(j) < maxRate)

       ! Check nearest neighbour distance
       if(ok) ok=(nnDist(srcExpIndex(j)) > minSeperation)

       ! Check source counts
       if(ok) ok=(srcCounts(srcExpIndex(j)) > minCounts)

       ! Passed all tests, so add to output source list
       if(ok) then
          nOutSrc=nOutSrc+1
          tempID(nOutSrc)=obsID(j)
          tempRA(nOutSrc)=srcRA(srcExpIndex(j))
          tempDec(nOutSrc)=srcDec(srcExpIndex(j))
          tempRate(nOutSrc)=countRate(j)
          tempCounts(nOutSrc)=srcCounts(srcExpIndex(j))
          tempDist(nOutSrc)=nnDist(srcExpIndex(j))
          flag(j)=.true.
       end if

       ! Decrement index
       i=i-1
       more=(i>=0 .and. nOutSrc<maxSources)
    end do

    ! Release exposure list dataset
    call release(expIDCol)
    call release(srcCtCol)
    call release(srcRACol)
    call release(srcDecCol)
    call release(instCol)
    call release(bandCol)
    call release(nnCol)
    call release(expTable)
    call release(expSet)

    ! Release observation list dataset
    call release(obsIDCol)
    call release(obsRateCol)
    call release(listXIDCol)
    call release(flagCol)
    call release(obsTable)
    call release(obsSet)

    ! Create output dataset
    outSet=dataset(outList, Create)
    outTable=addTable(outSet, 'SRCLIST', nOutSrc+1)
    outIDCol=addColumn(outTable, 'SRC_NUM', Integer32)
    outRACol=addColumn(outTable, 'RA', Real64)
    call setUnits(outRACol, 'degree')
    outDecCol=addColumn(outTable, 'DEC', Real64)
    call setUnits(outDecCol, 'degree')
    outRateCol=addColumn(outTable, 'RATE', Real32)
    call setUnits(outRateCol, 'ct/sec')
    outCountsCol=addColumn(outTable, 'COUNTS', Real32)
    call setUnits(outCountsCol, 'ct')
    outDistCol=addColumn(outTable, 'DIST_NN', Real32)
    call setUnits(outDistCol, 'arcsec')

    ! Write number of sources as an attribute
    call setAttribute(outSet, 'NSRC', nOutSrc)
    call setAttribute(outSet, 'MINCTS', minCounts)
    call setAttribute(outSet, 'MINRATE', minRate)
    call setAttribute(outSet, 'MAXRATE', maxRate)
    call setAttribute(outSet, 'MINSEP', minSeperation)
    call setAttribute(outSet, 'MAXSRC', maxSources)

    ! Populate table columns
    outID=>int32Data(outIDCol)
    outRA=>real64Data(outRACol)
    outDec=>real64Data(outDecCol)
    outRate=>real32Data(outRateCol)
    outCounts=>real32Data(outCountsCol)
    outDist=>real32Data(outDistCol)

    outID=tempID(0:nOutSrc)
    outRA=tempRA(0:nOutSrc)
    outDec=tempDec(0:nOutSrc)
    outCounts=tempCounts(0:nOutSrc)
    outRate=tempRate(0:nOutSrc)
    outDist=tempDist(0:nOutSrc)

    ! Free memory
    deallocate(tempID)
    deallocate(tempDec)
    deallocate(tempRA)
    deallocate(tempCounts)
    deallocate(tempRate)
    deallocate(tempDist)

    ! Release output dataset
    call release(outIDCol)
    call release(outRACol)
    call release(outDecCol)
    call release(outRateCol)
    call release(outCountsCol)
    call release(outDistCol)
    call release(outTable)
    call release(outSet)

  end subroutine esrcselect
  
end module esrcselect_mod
